/*
 *  linux/fs/nfs/rpcsock.c
 *
 *  This is a generic RPC call interface for datagram sockets that is able
 *  to place several concurrent RPC requests at the same time. It works like
 *  this:
 *
 *  -	When a process places a call, it allocates a request slot if
 *	one is available. Otherwise, it sleeps on the backlog queue
 *	(rpc_reserve).
 *  -	Then, the message is transmitted via rpc_send (exported by name of
 *	rpc_transmit).
 *  -	Finally, the process waits for the call to complete (rpc_call):
 *	The first process on the receive queue waits for the next RPC packet,
 *	and peeks at the XID. If it finds a matching request, it receives
 *	the datagram on behalf of that process and wakes it up. Otherwise,
 *	the datagram is discarded.
 *  -	If the process having received the datagram was the first one on
 *	the receive queue, it wakes up the next one to listen for replies.
 *  -	It then removes itself from the request queue (rpc_release).
 *	If there are more callers waiting on the backlog queue, they are
 *	woken up, too.
 *
 * Mar 1996:
 *  -	Split up large functions into smaller chunks as per Linus' coding
 *	style. Found an interesting bug this way, too.
 *  -	Added entry points for nfsiod.
 *
 *  Copyright (C) 1995, 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/nfs_fs.h>
#include <linux/errno.h>
#include <linux/socket.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/net.h>
#include <linux/mm.h>
#include <linux/rpcsock.h>

#include <linux/udp.h>
#include <net/sock.h>

#include <asm/segment.h>

#define msleep(sec)	{ current->timeout = sec * HZ / 1000; \
			  current->state = TASK_INTERRUPTIBLE; \
			  schedule(); \
			}

#undef DEBUG_RPC	/* debug tracing */
#define PROFILE_RPC	/* profiling */
#define SOCK_HAS_USER_DATA

#ifdef DEBUG_RPC			
static int			dodebug = 0;
#define dprintk(args...)	if (dodebug) printk(## args)
#else
#define	dprintk(args...)
#endif

#ifdef PROFILE_RPC
# define pprintk(args...)	printk(## args)
#else
# define pprintk(args...)
#endif


/*
 * Local variables
 */
static struct rpc_clnt *	sock_ready = NULL;
static struct rpc_clnt *	sock_pending = NULL;
static struct rpc_clnt *	sock_dying = NULL;
static struct rpc_wait *	async_pending = NULL;
#ifndef SOCK_HAS_USER_DATA
static struct rpc_clnt *	sock_list = NULL;
#endif
static struct wait_queue *	rpciod_idle = NULL;
int				rpciod_pid = 0;

/*
 * Append socket to queue (circular list)
 */
static inline void
rpc_clnt_enqueue(struct rpc_clnt **q, struct rpc_clnt *clnt)
{
	struct rpc_clnt	*prev, *next;

	dprintk("RPC: enqueuing client %p on %p (head %p)\n", clnt, q, *q);
	if (!(next = *q)) {
		*q = clnt->next = clnt->prev = clnt;
	} else {
		prev = (*q)->prev;
		clnt->prev = prev;
		clnt->next = next;
		prev->next = clnt;
		next->prev = clnt;
	}
}

/*
 * Remove socket from queue
 */
static inline void
rpc_clnt_dequeue(struct rpc_clnt **q, struct rpc_clnt *clnt)
{
	struct rpc_clnt	*next = clnt->next,
			*prev = clnt->prev;

	dprintk("RPC: dequeuing client %p from %p (head %p)\n", clnt, q, *q);
	if (next == clnt) {
		*q = NULL;
	} else {
		next->prev = prev;
		prev->next = next;
		if (*q == clnt)
			*q = next;
	}
}


/*
 * Insert new request into wait list.
 * New code: make the list sorted by timeout to increase chances that
 * the next reply arriving is really for the first request on the list.
 * This reduces the number of context switches required upon data receipt
 * enormously.
 * FIXME: requires per-procedure RTO estimate for best effect.
 */
static inline void
rpc_slot_enqueue(struct rpc_clnt *clnt, struct rpc_wait *slot)
{
#if 0
	struct rpc_wait	*next = clnt->pending;

	dprintk("RPC: enqueuing slot %p on pending (head %p)\n",
						slot, clnt->pending);
	slot->w_next = next;
	slot->w_prev = NULL;
	if (next)
		next->w_prev = slot;
	clnt->pending = slot;
#else
	struct rpc_wait	**q = &clnt->pending, *next, *prev = NULL;

	dprintk("RPC: enqueuing slot %p on pending (head %p)\n",
						slot, clnt->pending);
	while ((next = *q) != NULL) {
		if (slot->w_expires < next->w_expires)
			break;
		prev = next;
		q = &next->w_next;
	}

	slot->w_next = next;
	slot->w_prev = prev;
	if (next)
		next->w_prev = slot;
	*q = slot;
#endif
	slot->w_queued = 1;
}

/*
 * Remove request from request queue
 */
static inline void
rpc_slot_dequeue(struct rpc_clnt *clnt, struct rpc_wait *slot)
{
	struct rpc_wait	*prev = slot->w_prev,
			*next = slot->w_next;

	dprintk("RPC: dequeuing slot %p from pending (head %p)\n",
						slot, clnt->pending);
	if (prev != NULL)
		prev->w_next = next;
	else
		clnt->pending = next;
	if (next != NULL)
		next->w_prev = prev;

	slot->w_queued = 0;
}

/*
 * Put slot on async wait list. We keep this sorted by increasing timeout
 * value
 */
static inline void
rpc_async_enqueue(struct rpc_wait **q, struct rpc_wait *slot)
{
	struct rpc_wait	*next, *prev = NULL;

	dprintk("RPC: enqueuing slot %p on async (expires %ld)\n",
						slot, slot->w_expires);
	while ((next = *q) != NULL) {
		if (slot->w_expires < next->w_expires)
			break;
		prev = next;
		q = &next->w_async_next;
	}

	slot->w_async_next = next;
	slot->w_async_prev = prev;
	if (next)
		next->w_async_prev = slot;
	*q = slot;
	slot->w_asyncq = 1;
}

static inline void
rpc_async_dequeue(struct rpc_wait **q, struct rpc_wait *slot)
{
	struct rpc_wait	*next = slot->w_async_next,
			*prev = slot->w_async_prev;

	if (next)
		next->w_async_prev = prev;
	if (prev)
		prev->w_async_next = next;
	else
		*q = next;
	slot->w_asyncq = 0;
	dprintk("RPC: dequeued slot %p from async\n", slot);
}

/*
 * Data ready callback
 */
static void
rpc_data_ready(struct sock *sk, int len)
{
	struct rpc_clnt	*clnt;

#ifdef SOCK_HAS_USER_DATA
	clnt = (struct rpc_clnt *) sk->user_data;
#else
	for (clnt = sock_list; clnt && sk != clnt->inet; clnt = clnt->link)
		;
	if (!clnt)
		return;
#endif
	dprintk("RPC: data_ready client %p\n", clnt);
	dprintk("RPC: data %d locked %d pending %p\n",
			clnt->data, clnt->locked, clnt->pending);
	if (!clnt->data++ && !clnt->locked) {
		rpc_clnt_dequeue(&sock_pending, clnt);
		rpc_clnt_enqueue(&sock_ready, clnt);
	}
	if (clnt->pending)
		wake_up(clnt->pending->w_waitp);
	clnt->rdtime = jiffies;

	/* XXX: call clnt->old_data_ready(sk, len); ??? */
}

/*
 * Write data to socket.
 */
static inline int
rpc_sendmsg(struct rpc_clnt *clnt, struct iovec *iov, int nr, int len,
				struct sockaddr *sap, int salen)
{
	struct socket	*sock = clnt->sock;
	struct msghdr	msg;
	unsigned long	oldfs;
	int		result;

	msg.msg_iov	= iov;
	msg.msg_iovlen	= nr;
	msg.msg_name	= sap;
	msg.msg_namelen = salen;
	msg.msg_accrights = NULL;

	oldfs = get_fs();
	set_fs(get_ds());
	result = sock->ops->sendmsg(sock, &msg, len, 0, 0);
	set_fs(oldfs);

	dprintk("RPC: rpc_sendmsg(iov %p, len %d) = %d\n", iov, len, result);
	return result;
}

/*
 * Read data from socket
 */
static inline int
rpc_recvmsg(struct rpc_clnt *clnt, struct iovec *iov,
			int nr, int len, int flags)
{
	struct socket	*sock = clnt->sock;
	struct sockaddr	sa;
	struct msghdr	msg;
	unsigned long	oldfs;
	int		result, alen;

	msg.msg_iov	= iov;
	msg.msg_iovlen	= nr;
	msg.msg_name	= &sa;
	msg.msg_namelen = sizeof(sa);
	msg.msg_accrights = NULL;

	oldfs = get_fs();
	set_fs(get_ds());
	result = sock->ops->recvmsg(sock, &msg, len, 1, flags, &alen);
	set_fs(oldfs);

	if (!(flags & MSG_PEEK)) {
#if 0
		if (result == -EAGAIN)
			clnt->data = 0;
		else if (result > 0)
			clnt->data--;
#else
		if (result > 0)
			clnt->data--;
#endif
		if (clnt->data < 0) {
			printk("RPC: bummer! (rpc_recvmsg)\n");
			clnt->data = 0;
		}
	}

	dprintk("RPC: rpc_recvmsg(iov %p, len %d) = %d; qlen = %d\n",
				iov, len, result, clnt->data);
	return result;
}

/*
 * Reserve an RPC call slot. expires determines whether we wait in case
 * of congestion or not, and for how long.
 */
int
rpc_reserve(struct rpc_ioreq *req, unsigned long expires)
{
	struct rpc_clnt	*clnt = req->rq_client;
	struct rpc_wait	*slot;

	req->rq_slot = NULL;

	while (clnt->cong > clnt->cwnd) {
		if (!expires)
			return -ENOBUFS;
		dprintk("RPC: rpc_reserve waiting on backlog\n");
		current->timeout = expires;
		interruptible_sleep_on(&clnt->backlog);
		if (current->timeout == 0)
			return -ETIMEDOUT;
		current->timeout = 0;
		if (current->signal & ~current->blocked)
			return -ERESTARTSYS;
		if (clnt->shutdown)
			return -EIO;
	}

	if (!(slot = clnt->free)) {
		printk(KERN_WARNING "RPC: free list inconsistency\n");
		return -ENOBUFS;
	}

	clnt->free = slot->w_next;
	clnt->cong += RPC_CWNDSCALE;	/* bump congestion value */

	slot->w_handler = NULL;
	slot->w_waitp = &slot->w_wait;
	slot->w_asyncq = 0;
	slot->w_queued = 0;
	slot->w_gotit = 0;
	slot->w_req = req;

	dprintk("RPC: reserved slot %p\n", slot);
	req->rq_slot = slot;
	return 0;
}

/*
 * Release an RPC call slot
 */
void
rpc_release(struct rpc_ioreq *req)
{
	struct rpc_clnt	*clnt = req->rq_client;
	struct rpc_wait	*slot = req->rq_slot;

	if (slot != NULL) {
		dprintk("RPC: release slot %p\n", slot);

		/* remove slot from queue of pending */
		if (slot->w_queued)
			rpc_slot_dequeue(clnt, slot);
		if (slot->w_asyncq)
			rpc_async_dequeue(&async_pending, slot);
		slot->w_next = clnt->free;
		clnt->free = slot;

		/* Wake up next receiver */
		if (clnt->data && clnt->pending && !slot->w_handler)
			wake_up(clnt->pending->w_waitp);

		/* decrease congestion value */
		clnt->cong -= RPC_CWNDSCALE;
		if (clnt->cong < clnt->cwnd && clnt->backlog)
			wake_up(&clnt->backlog);
		if (clnt->shutdown)
			wake_up(&clnt->shutwait);

		req->rq_slot = NULL;
	}
}

/*
 * Adjust RPC congestion window
 */
static void
rpc_adjust_cwnd(struct rpc_clnt *clnt, int result)
{
	unsigned long	cwnd = clnt->cwnd;

	if (result >= 0) {
		if (clnt->cong >= cwnd) {
			/* The (cwnd >> 1) term makes sure
			 * the result gets rounded properly. */
			cwnd += (RPC_CWNDSCALE * RPC_CWNDSCALE +
					(cwnd >> 1)) / cwnd;
			if (cwnd > RPC_MAXCWND)
				cwnd = RPC_MAXCWND;
			else
				pprintk("RPC: %lu %ld cwnd\n", jiffies, cwnd);
			dprintk("RPC: cong %08lx, cwnd was %08lx, now %08lx\n",
				clnt->cong, clnt->cwnd, cwnd);
		}
	} else if (result == -ETIMEDOUT) {
		if ((cwnd >>= 1) < RPC_CWNDSCALE)
			cwnd = RPC_CWNDSCALE;
		dprintk("RPC: cong %08lx, cwnd was %08lx, now %08lx\n",
				clnt->cong, clnt->cwnd, cwnd);
		pprintk("RPC: %lu %ld cwnd\n", jiffies, cwnd);
	}

	clnt->cwnd = cwnd;
}

/*
 * Adjust timeout values etc for next retransmit
 */
static int
rpc_adjust_timeout(struct rpc_timeout *to)
{
	if (to->to_exponential)
		to->to_current <<= 1;
	else
		to->to_current += to->to_increment;
	if (to->to_maxval && to->to_current >= to->to_maxval)
		to->to_current = to->to_maxval;
	if (!to->to_current) {
		printk(KERN_WARNING "rpc_adjust_timeout: to_current = 0!\n");
		to->to_current = 5 * HZ;
	}
#ifdef PROFILE_RPC
	if (!to->to_retries || to->to_retries)
		printk("RPC: %lu retrans\n", jiffies);
#endif
	return (to->to_retries)--;
}

/*
 * Place the actual RPC call.
 * We have to copy the iovec because sendmsg fiddles with its contents.
 */
static inline int
rpc_send(struct rpc_clnt *clnt, struct rpc_wait *slot)
{
	struct rpc_ioreq *req = slot->w_req;
	struct iovec	iov[MAX_IOVEC];

	if (clnt->shutdown)
		return -EIO;

	memcpy(iov, req->rq_svec, req->rq_snr * sizeof(iov[0]));
	slot->w_xid = *(u32 *)(iov[0].iov_base);
	if (!slot->w_queued)
		rpc_slot_enqueue(clnt, slot);

	dprintk("RPC: rpc_send(%p, %x)\n", slot, slot->w_xid);
	return rpc_sendmsg(clnt, iov, req->rq_snr, req->rq_slen,
					&clnt->addr, clnt->alen);
}

/*
 * Receive and dispatch a single reply. This the ugliest part of it all.
 */
static inline void
rpc_grok(struct rpc_clnt *clnt)
{
	struct rpc_wait	*rovr;
	struct rpc_ioreq *req;
	struct iovec	iov[MAX_IOVEC];
	u32		xid;
	int		safe, result;

	iov[0].iov_base = (void *) &xid;
	iov[0].iov_len  = sizeof(xid);
	result = rpc_recvmsg(clnt, iov, 1, sizeof(xid), MSG_PEEK);

	if (result < 0) {
		if (result != -EAGAIN && result != ECONNREFUSED)
			dprintk("rpc_grok: recv error = %d\n", result);
		return;
	}
	if (result < 4) {
		printk(KERN_WARNING "RPC: impossible RPC reply size %d\n",
						result);
		return;
	}

	dprintk("RPC: rpc_grok: got xid %08lx\n", (unsigned long) xid);

	/* Look for the caller */
	safe = 0;
	for (rovr = clnt->pending; rovr; rovr = rovr->w_next) {
		if (rovr->w_xid == xid)
			break;
		if (safe++ > RPC_MAXREQS) {
			printk(KERN_WARNING "RPC: loop in request Q!!\n");
			rovr = NULL;
			break;
		}
	}

	if (!rovr || rovr->w_gotit) {
		/* discard dgram */
		dprintk("RPC: rpc_grok: %s.\n",
			rovr? "duplicate reply" : "bad XID");
		iov[0].iov_base = (void *) &xid;
		iov[0].iov_len  = sizeof(xid);
		rpc_recvmsg(clnt, iov, 1, sizeof(xid), 0);
		return;
	}
	req = rovr->w_req;

	/* Okay, we have it. Lock it to make sure nothing bad happens
	 * to us if recvmsg should sleep */
	rovr->w_locked = 1;

	/* Now receive the reply... Copy the iovec first because of 
	 * memcpy_fromiovec fiddling. */
	memcpy(iov, req->rq_rvec, req->rq_rnr * sizeof(iov[0]));
	if ((result = rpc_recvmsg(clnt, iov, req->rq_rnr, req->rq_rlen, 0)) < 0)
		return;

	req->rq_rlen   = result;
	rovr->w_result = result;
	rovr->w_gotit  = 1;
	rovr->w_locked = 0;
	if (rovr->w_asyncq)
		rpc_async_dequeue(&async_pending, rovr);

#ifdef PROFILE_RPC
	/* Profile only reads for now */
	if (result > 1024) {
		static unsigned long	nextstat = 0;
		static unsigned long	pkt_rtt = 0, pkt_len = 0, pkt_cnt = 0;

		pkt_cnt++;
		pkt_len += req->rq_slen + result;
		pkt_rtt += jiffies - req->rq_xtime;
		if (nextstat < jiffies) {
			printk("RPC: %lu %ld cwnd\n", jiffies, clnt->cwnd);
			printk("RPC: %ld %ld %ld %ld stat\n",
					jiffies, pkt_cnt, pkt_len, pkt_rtt);
			pkt_rtt = pkt_len = pkt_cnt = 0;
			nextstat = jiffies + 5 * HZ;
		}
	}
#endif

	/* ... and wake up the process. */
	if (rovr->w_handler) {
		dprintk("RPC: kicking callback handler %p\n",
				rovr->w_handler);
		rpc_adjust_cwnd(clnt, result);
		rovr->w_handler(result, rovr->w_req);
	} else {
		wake_up(rovr->w_waitp);
	}

	return;
}

static inline int
rpc_dispatch(struct rpc_clnt *clnt)
{
	unsigned long	oldflags;

	save_flags(oldflags); cli();
	if (clnt->locked || !clnt->data) {
		restore_flags(oldflags);
		return 0;
	}
	clnt->locked = 1;
	/* If you see this message, this is a bug. */
	if (jiffies > clnt->rdtime + 30) {
		printk("RPC: %ld %d %ld stuckdata\n", jiffies,
			clnt->data, jiffies - clnt->rdtime);
	}
	rpc_clnt_dequeue(&sock_ready, clnt);
	sti();	/* re-enable interrupts during rpc_grok */
	rpc_grok(clnt);
	cli();

	if (!clnt->data) {
		rpc_clnt_enqueue(&sock_pending, clnt);
	} else {
		rpc_clnt_enqueue(&sock_ready, clnt);
		clnt->rdtime = jiffies;
	}
	clnt->locked = 0;
	restore_flags(oldflags);
	return 1;
}

/*
 * Wait for the reply to our call.
 */
static int
rpc_recv(struct rpc_clnt *clnt, struct rpc_wait *slot)
{
	while (!slot->w_gotit) {
		while (!slot->w_gotit && rpc_dispatch(clnt))
			;
		if (!slot->w_gotit) {
			current->timeout = slot->w_expires;
			interruptible_sleep_on(slot->w_waitp);
			/* go back to sleep if someone is receiving data
			 * for us */
			if (slot->w_locked)
				continue; 
			if (current->timeout == 0)
				return -ETIMEDOUT;
			current->timeout = 0;
		}
		if (current->signal & ~current->blocked)
			return -ERESTARTSYS;
		if (clnt->shutdown)
			return -EIO;
	}

	return slot->w_result;
}

/*
 * Initialize an RPC I/O request
 * The cred parameter is not yet used, but intended for future extensions.
 */
void
rpc_initreq(struct rpc_clnt *clnt, struct rpc_ioreq *req)
{
	memset(req, 0, sizeof(*req));
	req->rq_timeout = clnt->timeout;
	req->rq_client = clnt;
}

/*
 * Generic RPC call routine. This handles retries and timeouts etc pp.
 */
int
rpc_call(struct rpc_ioreq *req)
{
	struct rpc_timeout *to = &req->rq_timeout;
	struct rpc_clnt	*clnt = req->rq_client;
	struct rpc_wait	*slot = NULL;
	int		result;

	do {
		dprintk("RPC: rpc_call: TP1 (req %p)\n", req);
		if (slot == NULL) {
			result = rpc_reserve(req, jiffies + to->to_current);
			if (result == -ETIMEDOUT)
				goto timedout;
			if (result < 0)
				break;
			slot = req->rq_slot;
		}

		dprintk("RPC: rpc_call: TP2\n");
		req->rq_xtime = jiffies;
		slot->w_expires = jiffies + to->to_current;
		if ((result = rpc_send(clnt, slot)) >= 0) {
			result = rpc_recv(clnt, slot);
			rpc_adjust_cwnd(clnt, result);
		}

		if (result != -ETIMEDOUT)
			break;
timedout:
		dprintk("RPC: rpc_recv returned timeout.\n");
		if (!rpc_adjust_timeout(to))
			break;
	} while (1);

	dprintk("RPC: rpc_call: TP3\n");
	rpc_release(req);
	return result;
}

/*
 * This one transmits an RPC call and leaves it to rpciod to wait for
 * the reply.
 */
int
rpciod_call(struct rpc_ioreq *req)
{
	struct rpc_clnt	*clnt = req->rq_client;
	struct rpc_wait	*slot = req->rq_slot;

	if (!rpciod_pid)
		return -ETIMEDOUT;

	slot->w_expires = jiffies + req->rq_timeout.to_current;
	slot->w_waitp   = &rpciod_idle;
	rpc_async_enqueue(&async_pending, slot);
	req->rq_xtime = jiffies;
	return rpc_send(clnt, slot);
}

/*
 * Loop over all stale async requests and retransmit them if appropriate
 */
static inline void
rpciod_retransmit(void)
{
	struct rpc_wait	*slot;
	struct rpc_ioreq *req;
	int		rexmit;

	dprintk("RPC: rpciod_retransmit\n");
	while ((slot = async_pending) && slot->w_expires <= jiffies) {
		rpc_async_dequeue(&async_pending, slot);
		if (slot->w_locked) {
			/* back off a little */
			dprintk("rpciod_retrans backoff %p\n", slot);
			slot->w_expires = jiffies + HZ / 10;
			rpc_async_enqueue(&async_pending, slot);
			continue;
		}
		req = slot->w_req;
		if (req->rq_client->data) {
			struct rpc_clnt	*clnt = req->rq_client;
			printk("RPC: %d %ld stuckdata\n",
				clnt->data, jiffies - clnt->rdtime);
		}
#ifdef DEBUG_RPC
		if (dodebug) {
			struct rpc_clnt	*clnt = req->rq_client;

			dprintk("RPC: client %p %slocked data %d\n",
					clnt, clnt->locked? "" : "un",
					clnt->data);
		}
#endif
		rpc_adjust_cwnd(req->rq_client, -ETIMEDOUT);
		rexmit = rpc_adjust_timeout(&req->rq_timeout);
		if (rexmit) {
			dprintk("rpciod_retrans re-xmit %p\n", slot);
			slot->w_expires = jiffies + req->rq_timeout.to_current;
			rpc_async_enqueue(&async_pending, slot);
			req->rq_xtime = jiffies;
			rpc_send(req->rq_client, slot);
		} else {
			dprintk("rpciod_retrans timeout %p\n", slot);
			slot->w_handler(-ETIMEDOUT, req);
		}
	}
}

static void
rpciod_abort(struct rpc_clnt *clnt)
{
	struct rpc_wait	*slot;

	dprintk("RPC: rpciod aborting all calls for client %p\n", clnt);
	do {
		for (slot = clnt->pending; slot; slot = slot->w_next) {
			if (slot->w_asyncq) {
				dprintk("RPC: rpciod aborting on slot %p\n",
							slot);
				slot->w_handler(-EIO, slot->w_req);
				break;
			}
		}
	} while (slot);
}

/*
 * This is the rpciod process.
 * For modules, this is an awful hack because when mounting an NFS volume,
 * the mount process's memory will stick around the swap space until the
 * volume is unmounted again.
 */
int
rpciod(void *ptr)
{
	struct wait_queue **assassin = (struct wait_queue **) ptr;
	unsigned long	oldflags;
	struct rpc_clnt	*clnt;
	struct rpc_wait	*slot;

#ifdef __SMP__
	lock_kernel();
	syscall_count++;
#endif

	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
	sprintf(current->comm, "rpciod");

	dprintk("RPC: rpciod starting up\n");
	rpciod_pid = current->pid;
	while (!(current->signal & ~current->blocked)) {
		/* dispatch replies */
		while ((clnt = sock_ready) != NULL)
			rpc_dispatch(clnt);

		/* terminate requests on dying client */
		while ((clnt = sock_dying) != NULL)
			rpciod_abort(clnt);

		/* Transmit timed-out requests */
		rpciod_retransmit();

		save_flags(oldflags); cli();
		if ((slot = async_pending) != NULL)
			current->timeout = slot->w_expires;
		interruptible_sleep_on(&rpciod_idle);
		restore_flags(oldflags);
		current->timeout = 0;
#ifdef DEBUG_RPC
		if (current->signal & (1<<(SIGUSR1-1))) {
			current->signal &= ~(1<<(SIGUSR1-1));
			dodebug = !dodebug;
			printk("RPC: debugging set to %d\n", dodebug);
		}
#endif
	}

	dprintk("RPC: rpciod terminating\n");
	rpciod_pid = 0;

	while ((slot = async_pending) != NULL) {
		rpciod_abort(slot->w_req->rq_client);
		/* play safe */
		if (slot == async_pending) {
			printk(KERN_WARNING "bummer (rpciod abort)\n");
			break;
		}
	}

	wake_up(assassin);
	return 0;
}

/*
 * Create and initialize an RPC client
 */
struct rpc_clnt *
rpc_create_client(struct file *file, struct sockaddr *ap, int alen,
					struct rpc_timeout *to)
{
	struct rpc_clnt	*clnt;
	struct socket	*sock;
	struct sock	*sk;
	struct rpc_wait	*slot;
	int		i;

	dprintk("RPC: make RPC socket...\n");
	sock = &file->f_inode->u.socket_i;
	if (sock->type != SOCK_DGRAM || sock->ops->family != AF_INET) {
		printk(KERN_WARNING "RPC: only UDP sockets supported\n");
		return NULL;
	}
	sk = (struct sock *) sock->data;

	if ((clnt = kmalloc(sizeof(struct rpc_clnt), GFP_KERNEL)) == NULL)
		return NULL;
	memset(clnt, 0, sizeof(*clnt)); /* Nnnngh! */

	clnt->sock = sock;
	clnt->inet = sk;
	clnt->file = file;
	clnt->cwnd = RPC_INITCWND;
	clnt->addr = *ap;
	clnt->alen = alen;
#ifdef SOCK_HAS_USER_DATA
	sk->user_data = clnt;
#else
	clnt->link = sock_list;
	sock_list = clnt;
#endif
	clnt->old_data_ready = sk->data_ready;
	sk->data_ready = rpc_data_ready;

	/* Set timeout parameters */
	clnt->timeout = *to;
	clnt->timeout.to_current = to->to_initval;

	dprintk("RPC: slots %p, %p, ...\n", clnt->waiting, clnt->waiting + 1);
	clnt->free = clnt->waiting;
	for (i = 0, slot = clnt->waiting; i < RPC_MAXREQS-1; i++, slot++)
		slot->w_next = slot + 1;
	slot->w_next = NULL;

	dprintk("RPC: made socket %p\n", clnt);
	rpc_clnt_enqueue(&sock_pending, clnt);
	return clnt;
}

int
rpc_destroy_client(struct rpc_clnt *clnt)
{
	unsigned long	t0 = jiffies;
#ifndef SOCK_HAS_USER_DATA
	struct rpc_clnt	**q;

	for (q = &sock_list; *q && *q != clnt; q = &((*q)->link))
		;
	if (!*q) {
		printk(KERN_WARNING "rpc_destroy_client: unknown socket!\n");
		return -EIO;	/* why is there no EBUGGYSOFTWARE */
	}
	*q = clnt->link;
#else
	clnt->inet->user_data = NULL;
#endif
	clnt->inet->data_ready = clnt->old_data_ready;

	rpc_clnt_dequeue(clnt->data? &sock_ready : &sock_pending, clnt);
	rpc_clnt_enqueue(&sock_dying, clnt);

	clnt->shutdown = 1;
	while (clnt->pending || clnt->backlog) {
		interruptible_sleep_on(&clnt->shutwait);
		if (current->signal & ~current->blocked)
			return -EINTR;
#if 1
		if (t0 && t0 - jiffies > 60 * HZ) {
			printk(KERN_WARNING
				"RPC: hanging in rpc_destroy_client.\n");
			t0 = 0;
		}
#endif
	}

	rpc_clnt_dequeue(&sock_dying, clnt);
	kfree(clnt);
	return 0;
}
