/*
 *  rpcsock.h	Declarations for the RPC call interface.
 *
 *  Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */


#ifndef _LINUX_RPCSOCK_H
#define _LINUX_RPCSOCK_H

/*
 * The rpcsock code maintains an estimate on the maximum number of out-
 * standing RPC requests, using the congestion avoidance implemented in
 * 44BSD. This is basically the Van Jacobson slow start algorithm: If a
 * retransmit occurs, the congestion window is halved; otherwise, it is
 * incremented by 1/cwnd when a reply is received and a full number of
 * requests are outstanding.
 *
 * Upper procedures may check whether a request would block waiting for
 * a free RPC slot by using the RPC_CONGESTED() macro.
 *
 * Note: on machines with low memory we should probably use a smaller
 * MAXREQS value: At 32 outstanding reqs with 8 megs of RAM, fragment
 * reassembly will frequently run out of memory.
 */
#define RPC_MAXREQS		32
#define RPC_MAXCONG		12
#define RPC_CWNDSCALE		256
#define RPC_MAXCWND		(RPC_MAXCONG * RPC_CWNDSCALE)
#define RPC_INITCWND		RPC_CWNDSCALE
#define RPC_CONGESTED(rsock)	((rsock)->cong >= (rsock)->cwnd)

/* RPC reply header size: xid, direction, status, accept_status (verifier
 * size computed separately)
 */
#define RPC_HDRSIZE		(4 * 4)

/*
 * This describes a timeout strategy
 */
struct rpc_timeout {
	unsigned long		to_current,		/* current timeout */
				to_initval,		/* initial timeout */
				to_maxval,		/* max timeout */
				to_increment;		/* if !exponential */
	int			to_retries;		/* max # of retries */
	char			to_exponential;
};

/*
 * This describes a complete RPC request
 */
struct rpc_ioreq {
	struct rpc_clnt *	rq_client;		/* RPC client */
	struct rpc_wait *	rq_slot;		/* RPC slot */
	struct rpc_timeout	rq_timeout;		/* timeout parms */
	struct iovec		rq_svec[MAX_IOVEC];	/* send buffer */
	unsigned int		rq_snr;
	unsigned long		rq_slen;
	struct iovec		rq_rvec[MAX_IOVEC];	/* recv buffer */
	unsigned int		rq_rnr;
	unsigned long		rq_rlen;

	/* profiling */
	unsigned long		rq_xtime;		/* when transmitted */

	void *			rq_calldata;		/* bio caller data */
};

/*
 * This is the callback handler for async RPC.
 */
typedef void	(*rpc_callback_fn_t)(int, struct rpc_ioreq *);

/*
 * Wait information. This struct defines all the state of an RPC
 * request currently in flight.
 */
struct rpc_wait {
	struct rpc_wait *	w_prev;		/* per-client request list */
	struct rpc_wait *	w_next;
	struct rpc_wait *	w_async_next;	/* async request list */
	struct rpc_wait *	w_async_prev;

	__u32			w_xid;		/* request XID */
	struct rpc_ioreq *	w_req;		/* I/O buffers */
	unsigned long		w_expires;	/* next re-xmit */
	int			w_result;	/* return value */
	struct wait_queue *	w_wait;		/* wait queue (sync reqs) */
	struct wait_queue **	w_waitp;	/* wait queue pointer */
	rpc_callback_fn_t	w_handler;	/* callback (async reqs) */
	unsigned int		w_queued : 1,	/* on client req list */
				w_asyncq : 1,	/* on async req list */
				w_locked : 1,	/* locked for receiving */
				w_gotit  : 1;	/* reply received */
};

struct rpc_clnt {
	struct rpc_clnt *	next;		/* ready list */
	struct rpc_clnt *	prev;
	struct rpc_clnt *	link;		/* list of all clients */

	struct file *		file;		/* VFS layer */
	struct socket *		sock;		/* BSD socket layer */
	struct sock *		inet;		/* INET layer */

	struct sockaddr		addr;		/* address */
	unsigned int		alen;
	struct rpc_timeout	timeout;	/* timeout parms */

	unsigned long		cong;		/* congestion control */
	unsigned long		cwnd;
	unsigned long		rdtime;		/* DBG: last data ready */
	int			data;		/* data ready */

	struct rpc_wait		waiting[RPC_MAXREQS];
	struct rpc_wait *	pending;
	struct rpc_wait *	free;
	struct wait_queue *	backlog;
	struct wait_queue *	shutwait;
	unsigned int		shutdown : 1,
				locked   : 1;

	void			(*old_data_ready)(struct sock *, int);
};

#ifdef __KERNEL__

struct rpc_clnt *	rpc_create_client(struct file *,
					struct sockaddr *, int,
					struct rpc_timeout *);
int			rpc_destroy_client(struct rpc_clnt *);

void			rpc_initreq(struct rpc_clnt *, struct rpc_ioreq *);
int			rpc_reserve(struct rpc_ioreq *, unsigned long timeout);
int			rpc_call(struct rpc_ioreq *);
int			rpc_transmit(struct rpc_ioreq *);
void			rpc_release(struct rpc_ioreq *);

/* Async RPC */
int			rpciod_call(struct rpc_ioreq *);
int			rpciod(void *dummy);
extern int		rpciod_pid;

#endif /* __KERNEL__*/

#endif /* _LINUX_RPCSOCK_H */
