/*
 *  linux/fs/nfs/sock.c
 *
 *  Copyright (C) 1992, 1993  Rick Sladkey
 *
 *  low-level nfs remote procedure call interface
 *
 * FIXES
 *
 * 2/7/94 James Bottomley and Jon Peatfield DAMTP, Cambridge University
 *
 * An xid mismatch no longer causes the request to be trashed.
 *
 * Peter Eriksson - incorrect XID used to confuse Linux
 * Florian La Roche - use the correct max size, if reading a packet and
 *                    also verify, if the whole packet has been read...
 *                    more checks should be done in proc.c...
 *
 */

#include <linux/sched.h>
#include <linux/nfs_fs.h>
#include <linux/errno.h>
#include <linux/socket.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/net.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/rpcsock.h>

#include <asm/segment.h>

#define _S(nr) (1<<((nr)-1))

/*
 * Place a synchronous call to the NFS server, meaning that the process
 * sleeps in rpc_call until it either receives a reply or a major timeout
 * occurs.
 */
int
nfs_rpc_call(struct nfs_server *server, int *start, int *end, int size)
{
	struct rpc_ioreq	req;
	unsigned long		maxtimeo;
	unsigned long		oldmask;
	int			major_timeout_seen, result;

	size += 1024;		/* account for NFS slack space. ugly */

	rpc_initreq(server->rsock, &req);
	req.rq_svec[0].iov_base = start;
	req.rq_svec[0].iov_len = (end - start) << 2;
	req.rq_slen = (end - start) << 2;
	req.rq_snr = 1;
	req.rq_rvec[0].iov_base = start;
	req.rq_rvec[0].iov_len = size;
	req.rq_rlen = size;
	req.rq_rnr = 1;

	oldmask = current->blocked;
	current->blocked |= ~(_S(SIGKILL)
		| ((server->flags & NFS_MOUNT_INTR)
		? ((current->sig->action[SIGINT - 1].sa_handler == SIG_DFL
			? _S(SIGINT) : 0)
		| (current->sig->action[SIGQUIT - 1].sa_handler == SIG_DFL
			? _S(SIGQUIT) : 0))
		: 0));

	major_timeout_seen = 0;
	maxtimeo = req.rq_timeout.to_maxval;

	do {
		result = rpc_call(&req);

		if (current->signal & ~current->blocked)
			result = -ERESTARTSYS;
		if (result == -ETIMEDOUT) {
			if (server->flags & NFS_MOUNT_SOFT) {
				printk("NFS server %s not responding, "
					"timed out.\n", server->hostname);
				result = -EIO;
				break;
			}
			if (!major_timeout_seen) {
				printk("NFS server %s not responding, "
					"still trying.\n", server->hostname);
				major_timeout_seen = 1;
			}
			if ((req.rq_timeout.to_initval <<= 1) >= maxtimeo)
				req.rq_timeout.to_initval = maxtimeo;
		} else if (result < 0 && result != -ERESTARTSYS) {
			printk("NFS: notice message: result = %d.\n", result);
		}
	} while (result == -ETIMEDOUT && !(server->flags & NFS_MOUNT_SOFT));

	if (result >= 0 && major_timeout_seen)
		printk("NFS server %s OK.\n", server->hostname);
	/* 20 is the minimum RPC reply header size */
	if (result >= 0 && result < 20) {
		printk("NFS: too small read memory size (%d bytes)\n", result);
		result = -EIO;
	}

	current->blocked = oldmask;
	return result;
}

struct rpc_ioreq *
nfs_async_reserve(struct nfs_server *server, rpc_callback_fn_t callback,
			void *clientdata)
{
	struct rpc_ioreq	*req;
	int			result;

	if (!rpciod_pid
	 || !(req = (struct rpc_ioreq *) kmalloc(sizeof(*req), GFP_NFS)))
		return NULL;

	rpc_initreq(server->rsock, req);
	if ((result = rpc_reserve(req, 0)) < 0) {
		kfree(req);
		return NULL;
	}

	req->rq_slot->w_handler = callback;
	req->rq_calldata = clientdata;
	return req;
}

void
nfs_async_release(struct rpc_ioreq *req)
{
	rpc_release(req);
	kfree(req);
}

int
nfs_async_call(struct rpc_ioreq *req)
{
	int	result;

	if ((result = rpciod_call(req)) < 0) {
		nfs_async_release(req);
	}
	return result;
}

